/* Libraries */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

/* Shared files */
#include "messaging.h"
#include "console.h"
#include "logging.h"
#include "mounting.h"

/* External files */
#include "usplash.h"
#include "usplash_child.h"

#define DEFAULT_CON 1
#define UNPRIV_UID 0
#define LOOP_SLEEP_USEC 500 * 1000

int child_pid;
int fbfd = -1;

struct option {
	int die_on_chvt;
} options = { 0 };


/* INIT FUNCTIONS */
/* Used to clean up the environment */
void ushutdown(int exitstatus, int vt) {
	logme("Exiting...");
	/* Shut down the event server */
	close_se_server();
	
	/* Umount proc + tmpfs */
	umount_proc();
	umount_tmpfs();
	
	/* Close framebuffer */
	close(fbfd);
	
	/* Change vt */
	if (vt > 0)
		change_console(vt);
	else
		change_console(DEFAULT_CON);
	
	/* Exit parent with specified exitstatus */
	exit(exitstatus);
}

/* Used to bring up the environment */
void ustartup(int vt) {	
	logme("Starting...");
	/* Make sure we are running as root */
	if (getuid()) {
		logme("Must run as root! Exiting...");
		ushutdown(1, 0);
	}
	/* Make sure we have graphical framebuffer */
	if (!graphical_framebuffer_enabled()) {
		logme("We don't have graphical framebuffer!");
		ushutdown(1, 0);		
	}
	
	/* Get current framebuffer's file discriptor */
	fbfd = get_framebuffer_fd();
	if (fbfd < 0) {
		logme("Unable to open framebuffer!");
		ushutdown(1, 0);
	}
	
	/* Mount tmpfs */
	if (mount_tmpfs()) {
		logme("Unable to mount tmpfs!");
		ushutdown(1, 0);
	}
	
	/* Mount proc */
	if (mount_proc()) {
		logme("Unable to mount proc!");
		ushutdown(1, 0);
	}
	
	/* Check to make sure 'splash' is in the kernel command line */
	
	/* Setup the server */
	if (open_se_server(SOCKET_LOCATION, s_event_handler)) {
		logme("Unable to start event server!");
		ushutdown(1, 0);
	}
		
	/* Change vt */
	if (vt <= 0)
		vt = SPLASH_CON;
	
	if(change_console(vt)) {
		logme("Unable to change console");
		ushutdown(1, 0);
	}
}


/* CALLBACKS */
/* When child dies */
void on_child_exit(int status) {
	int exitstatus;
	
	exitstatus = WEXITSTATUS(status);
	logme("Child has exited with exitstatus %d", exitstatus);
	if (exitstatus >= 0)
		ushutdown(0, exitstatus);
	else
		ushutdown(1, 0);
}
/* When parent receives a signal */
void parent_signal_handler(int sig) {
	int status;
	
	/* Log it */
	logme("Parent received signal %d", sig);
	
	/* if the child has died, lets get the exitstatus */
	if (sig == SIGCHLD) {
		waitpid(child_pid, &status, 0);
		on_child_exit(status);
	}
	else if (sig == SIGTERM)
		kill(child_pid, SIGTERM);
}

/* PARENT POST FORK MAIN LOOP */
/* Parent stuff post fork */
int parent_main() {
	int child_status;
		
	/* Ignore the more common signals */
	signal(SIGTSTP,SIG_IGN);
	signal(SIGHUP,SIG_IGN);
		
	/* Find out if the child dies */
	signal(SIGCHLD,parent_signal_handler);
	signal(SIGTERM,parent_signal_handler);
		
	/* Main parent loop */
	while (1) {
		/* Sleep */
		usleep(LOOP_SLEEP_USEC);
		
		/* Make sure the child isn't dead, if it is, we die */
		if (waitpid(child_pid, &child_status, WNOHANG) > 0)
			on_child_exit(child_status);
		
		/* Make sure we haven't changed terminals */
		if (get_current_console() != SPLASH_CON && options.die_on_chvt)
			kill(child_pid, SIGTERM);
	}

	/* DONT GET HERE */
	logme("We went to the naughty place!");
	return 1;
}

usage(char *progname) {
	puts("FIXME: usplashd.c -- usage()");
}

int main (int argc, char **argv) {
	int i;
	
	set_logging_term(9);
	
	/* Get our user supplied options */	
	for (i = 1 ; i < argc ; i++) {
		if (!strcmp(argv[i], "--die-on-chvt"))
			options.die_on_chvt = 1;
		else
			usage(argv[0]);
			exit(1);
	}

	/* Startup */
	ustartup(SPLASH_CON);
	
	/* START FORK */
	child_pid = fork();
		/* CHILD PROCESS */
	if (child_pid == 0)
		return child_main();
		/* PARENT PROCESS */
	else if (child_pid > 0)
		return parent_main();
		

	/* FAILED FORK (We should not get here!!!) */
	logme("Failed to fork!!!");
	ushutdown(1, 0);
}
