#include <stdio.h> /* For printf calls */
#include <stdarg.h> /* For va_lists (printf emulation) */
#include <linux/fb.h> /* For getting framebuffer info */
#include <linux/kd.h> /* For KDGKBTYPE definition */
#include <linux/vt.h> /* For getting virtual terminal info */
#include <sys/ioctl.h> /* For ioctl */

/* For open() */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

/* Local Libraries */
#include "console.h"
#include "logging.h" /* For logme() */

#define FB_PREFIX "/dev/fb"
#define CON_PREFIX "/dev/tty"
#define BUFFER 1000

#define ENABLE_CURSOR "\e[?25h\e[?0c"
#define DISABLE_CURSOR "\e[?25l\e[?1c"
#define CLEAR_CONSOLE "\e[H\e[J"

#ifndef FB_TYPE_PACKED_PIXELS
enum {
	FB_TYPE_PACKED_PIXELS = 0,			/* Packed Pixels	*/
	FB_TYPE_PLANES = 1,					/* Non interleaved planes */
	FB_TYPE_INTERLEAVED_PLANES = 2,		/* Interleaved planes	*/
	FB_TYPE_TEXT = 3,					/* Text/attributes	*/
	FB_TYPE_VGA_PLANES = 4,				/* EGA/VGA planes	*/

	FB_VISUAL_MONO01 = 0,				/* Monochr. 1=Black 0=White */
	FB_VISUAL_MONO10 = 1,				/* Monochr. 1=White 0=Black */
	FB_VISUAL_TRUECOLOR = 2,			/* True color	*/
	FB_VISUAL_PSEUDOCOLOR = 3,			/* Pseudo color (like atari) */
	FB_VISUAL_DIRECTCOLOR = 4,			/* Direct color */
	FB_VISUAL_STATIC_PSEUDOCOLOR = 5	/* Pseudo color readonly */
};
#endif

/*
** Determines whether a file descriptor is a console or not
** RETURN non-zero on success
*/
int is_a_console(int fd) {
	char arg;
  
	arg = 0;
	return (ioctl(fd, KDGKBTYPE, &arg) == 0 && ((arg == KB_101) || (arg == KB_84)));
}

/*
** Returns the filediscriptor of the master virtual terminal
** or negative on fail.  Dont forget to close it!
** RETURN non-negative on success
*/
int get_master_console_fd () {
	int fd; 
	
	/* This one must go first or we'll get permission denied when in X */
	fd = open("/dev/tty0", O_RDWR | O_NDELAY);	
	if (fd >= 0)
		return fd;
	
	fd = open("/dev/tty", O_RDWR | O_NDELAY);
	if (fd >= 0 )
		return fd;

	fd = open("/dev/console", O_RDWR | O_NDELAY);  
	if (fd >= 0)
		return fd;
	
	for (fd = 0; fd < 3; fd++)
		if (is_a_console(fd))
			return fd;
		
	logme("Can't open console!");
	return -1;
}

/*
** Opens console number 'console' and returns the file descriptor 
*/
int get_console_fd(int console) {
	int fd;
	char tmpstr[20];
	
	snprintf(tmpstr, 20, "%s%d", CON_PREFIX, console);
	fd = open(tmpstr, O_RDWR);
	if (fd < 0)
		return -1;
	return fd;
}


/*
** Returns the integer of the current console or negative on error
** RETURN non-negative on success
*/
int get_current_console() {
	int fd;
	char tempint = 12;
	
	fd = get_master_console_fd();
	if (fd < 0)
		return -1;
	tempint = ioctl(fd, TIOCLINUX, &tempint) + 1;
	close(fd);
	return tempint;
}

/*
** Returns a filediscriptor for the current framebuffer device
** or negative on error.  Don't forget to close it!
** NOTE: The FD is opened r+w (is there a case when this is a problem? -PS)
** RETURN non-negative on success
*/
int get_framebuffer_fd() {
	int fd = -1;
	int i;
	char tempstr[20];
	
	for (i=0 ; fd < 0 && i < 12 ; i++) {
		sprintf(tempstr, "%s%d", FB_PREFIX, i);
		fd = open(tempstr,  O_RDWR);
	}
	
	return fd;
}

/*
** Checks to make sure we have a graphical framebuffer.
** Returns 1 if graphical *is* enabled, 0 if not enabled
** RETURN non-zero on success
*/
int graphical_framebuffer_enabled() {
	int fd;
	struct fb_fix_screeninfo info;
	
	fd = get_framebuffer_fd();
	if (fd < 0)
		return 0;
	if (ioctl(fd, FBIOGET_FSCREENINFO, &info) == -1) {
		close(fd);
		return 0;
	}
	close(fd);
	
	if (info.type != FB_TYPE_PACKED_PIXELS)
		return 0;
	switch (info.visual) {
		case FB_VISUAL_TRUECOLOR:
			return 1;
		case FB_VISUAL_PSEUDOCOLOR:
			return 1;
		default:
			return 0;
	}
	
	/* Don't get here */
	return 0;
}

/*
** Returns the length of the framebuffer memory or 0 for failure
** RETURN non-zero on success
*/
unsigned long get_framebuffer_length() {
	int fd;
	struct fb_fix_screeninfo info;
	
	fd = get_framebuffer_fd();
	if (fd < 0)
		return 0;
	if (ioctl(fd, FBIOGET_FSCREENINFO, &info) == -1) {
		close(fd);
		return 0;
	}
	close(fd);

	return info.smem_len;	
}

/*
** Returns a pointer to a (static) struct with info on the framebuffer. 
** RETURN non-zero on success
*/
const struct fb_var_screeninfo *get_framebuffer_specs() {
	static struct fb_var_screeninfo info;
	int fd;
	
	/* Get the info */
	fd = get_framebuffer_fd();
	if (fd < -1)
		return NULL;
	if (ioctl(fd, FBIOGET_VSCREENINFO, &info) == -1) {
		close(fd);	
		return NULL;
	}
	close(fd);
	
	return &info;
}

/*
** A printf implimentation that prints to the current number specified 
** rather than /dev/console
** RETURN non-negative on success
*/
int con_printf(int console, char *format, ...) {
	int fd;
	va_list ap;
	char string[BUFFER], con_s[12];
	int status;
	
	/* Get the string */
	va_start(ap, format);
	status = vsnprintf(string, BUFFER, format, ap);
	va_end(ap);
	if (status < 0)
		return -1;
	
	/* If we have an invalid console, just print to the current one */
	if (console < 0)
		return cur_con_printf(string);
	
	/* Open the console specified */
	fd = get_console_fd(console);
	if (fd < 0)
		return -1;

  	/* Write the string */
	status = write(fd, string, strlen(string));
	close(fd);
	if (status < 0)
		return -1;
	return status;
}

/*
** A printf implimentation that prints to the current console device 
** rather than /dev/console (which may be statically linked in the kernel)
*/
int cur_con_printf(char *format, ...) {
	va_list ap;
	int i;
	char string[BUFFER];

	/* Get the string */
	va_start(ap, format);
	i = vsnprintf(string, BUFFER, format, ap);
	va_end(ap);
	if (i < 0)
		return -1;
	
	/* Open the current console */
	if ((i = get_current_console()) < 0)
		return -1;
	return con_printf(i, string);
}

/*
** Changes to a specific terminal.  Returns 0 on success, nonzero on fail.
** RETURN zero on success
*/
int change_console(int console) {
	int master;
	
	master = get_master_console_fd();
	if (master < 0)
		return -1;
	
	if (ioctl(master, VT_ACTIVATE, console) || ioctl(master, VT_WAITACTIVE, console)) {
		close(master);
		 fprintf(stderr, "Can't change virtual terminal1 to %d!\n", console); 
		return -1;
	}
	close(master);
	return 0;
}

/*
** Turns off the cursor on the given terminal (use -1 for current terminal)
** RETURN non-negative on success
*/
int disable_cursor(int fd) {
	return con_printf(fd, DISABLE_CURSOR);
}
/*
** Turns on the cursor on the given terminal (use -1 for current terminal)
** RETURN non-negative on success
*/
int enable_cursor(int fd) {
	return con_printf(fd, ENABLE_CURSOR);
}

/*
** Clear text on the given terminal (use -1 for current terminal)
** RETURN non-negative on success
*/
int clear_console(int fd) {
	return con_printf(fd, CLEAR_CONSOLE);
}

void display_fbinfo(const struct fb_var_screeninfo *info)
{
  logme("found an FB device:\n"
	 "xres:\t\t%d\n"
	 "yres:\t\t%d\n"
	 "bpp:\t\t%d\n",
	 info->xres, info->yres, info->bits_per_pixel );
}

int get_framebuffer_size(const struct fb_var_screeninfo *info)
{
  return (info->xres * info->yres * info->bits_per_pixel) >> 3;
}
