#include <stdlib.h> /* For malloc */
#include <asm-generic/errno-base.h> /* For EAGAIN definition */
#include <string.h> /* For string functions (strlen(), etc) */

/* For fcntl() */
#include <unistd.h>
#include <fcntl.h>

/* For sockets */
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>

/* Local library */
#include "logging.h"
#include "messaging.h"

#define BUFFER 128
#define SERVER_QUEUE_LEN 10

static int server_socket = -1;
static struct sockaddr_un server_addr;
static void (*server_callback)(s_event *) = NULL;

/* Creates a s_event (must be freed when done) */
s_event *se_create(char *version, char *name, char *state, int value) {
	s_event *event;
	
	/* Allocate event */
	event = (s_event *)malloc(sizeof(s_event));
	if (!event)
		return NULL;
	event->version = (char *)malloc(strlen(version) + 1);
	if (!(event->version)) {
		free(event);
		return NULL;
	}
	event->name = (char *)malloc(strlen(name) + 1);
	if (!(event->name)) {
		free(event->version);
		free(event);
		return NULL;
	}
	event->state = (char *)malloc(strlen(state) + 1);
	if (!(event->state)) {
		free(event->version);
		free(event->name);
		free(event);
		return NULL;
	}
	
	/* Set data */
	sprintf(event->version, "%s", version);
	sprintf(event->name, "%s", name);
	sprintf(event->state, "%s", state);
	event->value = value;
	
	return event;
}

/* Frees a s_event */
void se_free(s_event *event) {
	free(event->version);
	free(event->name);
	free(event->state);
	free(event);
}

/* Returns the length of a s_event after being converted to a string */
int se_strlen(s_event *event) {
	int i;
	char tmpstr[20];
	
	/* Convert the int to a string */
	sprintf(tmpstr, "%d", event->value);
	
	/* Get the length of all the strings (+4 for colons and newline) */
	i = strlen(tmpstr) + strlen(event->version);
	i = i + strlen(event->name) + strlen(event->state) + 4;
	
	return i;
}

/* Turns an s_event into a string */
const char *se_to_string(s_event *event) {
	static char string[BUFFER];
	
	snprintf(string, BUFFER, EVENT_FORMAT_PSTRING, 
					event->version, event->name, event->state, event->value);
	
	return string;
}

/* Turns a string into an s_event, returns NULL on fail */
s_event *se_from_string(char *string) {
	char *version, *name, *state, *strarray[3];
	char *tmpstr, tmpstrp;
	int value, i = 0, j = 0;
	s_event *event;
	
	/* Allocate */
	version = (char *)malloc(strlen(string) + 1);
	if (!version)
		return NULL;
	version[0] = '\0';
	name = (char *)malloc(strlen(string) + 1);
	if (!name) {
		free(version);
		return NULL;
	}
	name[0] = '\0';
	state = (char *)malloc(strlen(string) + 1);
	if (!state) {
		free(version);
		free(name);
		return NULL;
	}
	state[0] = '\0';
	
	strarray[0] = version;
	strarray[1] = name;
	strarray[2] = state;

	/* Get our variables */
	for (i=0 ; i < 3 ; i++) {
		tmpstr = strchr(string, ':');
		if (!tmpstr) {
			free(version);
			free(name);
			free(state);
			return NULL;
		}
		strncat(strarray[i], string, tmpstr - string);
		string = (tmpstr + 1);
	}
	
	/* If we are here it means we sucessfully parsed the first 3 fields.
	Now we need to get the last field */
	value = atoi(string);

	/* Create event and return it */
	event = se_create(version, name, state, value);
	free(version);
	free(name);
	free(state);
	return event;
}


/* Sends an s_event to the specified location as a Unix Domain socket */
int se_send(s_event *event, char *us_location) {
	const char *message;
	int sock;
	struct sockaddr_un sock_addr;
	
	/* Convert s_event to string */
	message = se_to_string(event);
	if (!message)
		return -1;
	
	/* Set the location info */
	sock_addr.sun_family = AF_UNIX;
	strcpy(sock_addr.sun_path, us_location);
	
	/* Send our event */
	if ((sock = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
		logme("Error: Cannot open Socket!");
		return -1;
	}
	if (connect(sock, (struct sockaddr*)&sock_addr, sizeof(sock_addr))) {
		logme("Error: Cannot connect to socket!");
		close(sock);
		return -1;
	}
	if (send(sock, message, strlen(message), 0) != strlen(message)) {
		logme("Error: Could not send socket event!");
		shutdown(sock, SHUT_RDWR);
		close(sock);
		return -1;
	}
	
	/* Close everything */
	shutdown(sock, SHUT_RDWR);
	close(sock);
	
	return 0;
}

/* Closes (shuts down) the se_server */
void close_se_server() {
	/* Close any open servers */
	if (server_socket > 0) {
		shutdown(server_socket, SHUT_RDWR);
		close(server_socket);
	}
	if (server_addr.sun_path)
		remove(server_addr.sun_path);
	server_callback = NULL;
}

/* Sets up a server to listen for events, returns 0 on success, nonzero on fail */
int open_se_server(char *us_location, void (*callback)(s_event *)) {

	/* Close any open servers */
	close_se_server();
		
	/* Set the location info */
	server_addr.sun_family = AF_UNIX;
	strcpy(server_addr.sun_path, us_location);

	/* Must be world-rightable, otherwise if usquirt can't pass messages if called
	   as a non-root user; */
	umask(SOCKET_UMASK);

	/* Setup the server */
	if ((server_socket = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
		logme("Error: Cannot create socket!");
		return -1;
	}
	fcntl(server_socket, F_SETFL, O_NONBLOCK);
	if (bind(server_socket, (struct sockaddr*)&server_addr, sizeof(server_addr))) {
		logme("Error: Cannot bind socket to file!");
		close(server_socket);
		server_socket = -1;
		return -1;
	}
	if (listen(server_socket, SERVER_QUEUE_LEN)) {
		logme("Error: Cannot listen on socket!");
		close(server_socket);
		server_socket = -1;
		return -1;
	}
	
	server_callback = callback;
	return 0;
}
/* Used to poll the previously setup server for incoming events, if an event
is found, the callback is called with the event as a parameter.  Don't forget
to free the event in the callback when you are done with it!!! */
void poll_se_server() {
	int i, addr_len, buffer = BUFFER;
	char pad[2];
	char *string;
	s_event *event;
	
	/* Setup our pad */
	pad[0] = '\0';
	pad[1] = '\0';
	
	/* Allocate our initial string */
	string = (char *)malloc(sizeof(char) * BUFFER);
	if (!string)
		return;
	string[0] = '\0';
	
	/* Make sure there is a server running */
	if (server_socket < 0) {
		logme("Error: No server running!");
		free(string);
		return;
	}
	
	/* Check for and accept any connections */
	addr_len = sizeof(server_addr);
	i = accept(server_socket, (struct sockaddr*)&server_addr, &addr_len);
	if (i == EAGAIN || i < 0) {
		free(string);
		return;
	}
	logme("Connection accepted");

	/* Get our string */
	pad[0] = '\0';
	pad[1] = '\0';
	while (recv(i, pad, sizeof(char), 0) == 1) {
		if ((strlen(string) + 2) > buffer) {
			buffer+=BUFFER;
			if (!realloc(string, buffer)) {
				free(string);
				shutdown(i, SHUT_RDWR);
				close(i);
				return;
			}
		}
		strcat(string, pad);
		if (pad[0] == '\0')
			break;
		pad[0] = '\0';
	}

	/* Close the socket */
	shutdown(i, SHUT_RDWR);
	close(i);
	
	/* Convert the string to an event */
	event = se_from_string(string);
	free(string);
	if (!event) {
		logme("Error: Received invalid socket event!");
		return;
	}
	logme("Received socket event -- %s", se_to_string(event));
	
	/* Send the event to the callback */
	server_callback(event);
}
