#include <sys/mman.h> /* For mmap()/munmap() */
#include <linux/fb.h> /* For framebuffer access */
#include <stdlib.h> /* For malloc()/free() */

/* For open() and fstat() */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

/* Our libraries */
#include "upcx.h"
#include "console.h"
#include "uimage.h"

#ifndef NULL
#define NULL 0
#endif

static void *fb_mmap = NULL;

/*
** Allocates a new uimage 
*/
uimage *uimage_new(void) 
{
	uimage *image;
	
	/* Allocate and zero the uimage */
	image = (void *)malloc(sizeof(uimage));
	if(image)
		memset(image, 0, sizeof(uimage)); /* Don't use bzero, depreciated */

	return image;
}

/*
** Frees up an old image
*/
void uimage_free(uimage *image)
{
	/* Free our image */
	free(image->data);
	free(image->colourmap);
	free(image);
}

/*
** Copies the contents of the framebuffer into a uimage 
*/
int uimage_from_framebuffer_fast(uimage *fbimage)
{
	const struct fb_var_screeninfo *info;
	int fd, len;

	/* Get our framebuffer info */
	if((info = get_framebuffer_specs()) == NULL)
		return -1;

	/* Get the framebuffer specs */
	fbimage->width = info->xres;
	fbimage->height = info->yres;
	fbimage->depth = info->bits_per_pixel;
	fbimage->stride = info->xres * (info->bits_per_pixel >> 3);

	return 0;
}

int uimage_from_framebuffer(uimage *fbimage)
{
	const struct fb_var_screeninfo *info;
	int fd, len;

	/* Get our framebuffer info */
	if((info = get_framebuffer_specs()) == NULL)
		return -1;
	if ((len = get_framebuffer_length()) == 0)
		return -2;
	fd = get_framebuffer_fd();

	/* Allocate enough memory to hold the framebuffer */
	if((fbimage->data = (void *)malloc(len)) == NULL)
		return -3;
	
	/* Copy the framebuffer */
	read(fd, fbimage->data, len);
	close(fd);
	
	/* Get the framebuffer specs */
	fbimage->width = info->xres;
	fbimage->height = info->yres;
	fbimage->depth = info->bits_per_pixel;
	fbimage->stride = info->xres * (info->bits_per_pixel >> 3);

	return 0;
}

/*
** Creates a uimage from a pcx file
*/
int uimage_from_pcx(uimage *image, char *filename)
{
	int fd, file_length, height, width;
	void *colourmap_start;
	pcx *header;
	struct stat st;

	/* Open file */
	if((fd = open(filename, O_RDONLY)) < 0)
		return -1;
	
	/* Test file */
	if(is_pcx_suitable(fd))
		return -2;
	
	/* Get file size */
	if (fstat(fd, &st))
		return -3;
	file_length = st.st_size;
	
	/* mmap the pcx file into our header */
	header = mmap(NULL, file_length, PROT_READ, MAP_SHARED, fd, 0);
	if (header == MAP_FAILED)
		return -4;
	
	/* Get the width and height of the image */
	width = header->xmax - header->xmin + 1;
	height = header->ymax - header->ymin + 1;
	
	/* Allocate enough room for the data and colourmap*/
	image->data = (void *)malloc(width * height);
	if (!image->data) {
		munmap(header, file_length);
		image->data = NULL;
		return -5;
	}
	image->colourmap = (void *)malloc(UIMAGE_COLOUR_MAP_LENGTH);
	if (!image->colourmap) {
		munmap(header, file_length);
		free(image->data);
		image->data = NULL;
		image->colourmap = NULL;
		return -6;
	}
		
	/* Decode the data */
	raw_pcx_decode(image->data, header->data, width, height);

	/* Set our info */
	image->width = width;
	image->height = height;
	image->depth = header->bits_per_pixel;
	image->stride = image->width;
	colourmap_start = (void *)header + file_length - PCX_COLOUR_MAP_LENGTH + 1;
	memcpy(image->colourmap, colourmap_start, UIMAGE_COLOUR_MAP_LENGTH);
	
	/* Clean up */
	munmap(header, file_length);
	
	return close(fd);
}


static inline const int fade16_fast3(unsigned long c, int i)
{
  unsigned long a, b, d, aa, bb, dd;
  aa = a = 0xf800f800;
  bb = b = 0x07e007e0;
  dd = d = 0x001f001f;

  a &= c;
  b &= c;
  d &= c;
  a >>= 8;
  b >>= 5;
  a *= i;
  b *= i;
  d *= i;
  a &= aa;
  b >>= 3;
  d >>= 8;
  b &= bb;
  d &= dd;
  return a | b | d;
}

static inline const int fade16_fast2(unsigned long c, int i)
{
  unsigned long a, b, aa, bb;
  aa = a = 0xf81f07e0;
  bb = b = 0x07e0f81f;

  i += i >> 7;
  i >>= 3; /* now in range 0..64 */
  a &= c;
  b &= c;
  a >>= 5;
  b *= i;
  a *= i;
  b >>= 5;
  a &= aa;
  b &= bb;
  return a | b;
;
}

static inline const int fade16(int c, int i)
{
  return 
    ((((c & 0xf800) * i) >> 8) & 0xf800) |
    ((((c & 0x07e0) * i) >> 8) & 0x07e0) |
    ((((c & 0x001f) * i) >> 8) & 0x001f);
}

static inline const int mix16_fast3(unsigned long c, unsigned long d, int i)
{
  unsigned long a, b, e, aa, bb, ee, aaa, bbb, eee;
  aaa = aa = a = 0xf800f800;
  bbb = bb = b = 0x07e007e0;
  eee = ee = e = 0x001f001f;

  a &= c; aa &= d;
  b &= c; bb &= d;
  e &= c; ee &= d;
  a >>= 8; aa >>= 8;
  b >>= 3; bb >>= 3;
  e <<= 3; ee <<= 3;
  a *= i; aa *= (255-i);
  b *= i; bb *= (255-i);
  e *= i; ee *= (255-i);
  a += aa;
  b += bb;
  e += ee;
  a &= aaa;
  b >>= 5;
  e >>= 11;
  b &= bbb;
  e &= eee;
  return a | b | e;
}

static inline const int mix16_fast2(unsigned long c, unsigned long d, int i)
{
  unsigned long a, b, aa, bb, aaa, bbb;
  aaa = aa = a = 0xf81f07e0;
  bbb = bb = b = 0x07e0f81f;

  i = i >> 3;
  a &= c; aa &= d;
  b &= c; bb &= d;
  a >>= 5; aa >>= 5;
  a *= i; aa *= (31-i);
  b *= i; bb *= (31-i);
  a += aa;
  b += bb;
  a &= aaa;
  b >>= 5;
  b &= bbb;
  a |= b;
  //a += (a >> 5) & 0x08410841;
  a += 0x08410841;
  return a;
}

static inline const int mix16(int c, int d, int i)
{
  i += i>>7;
  return 
    (((((c & 0xf800) * i) + ((d & 0xf800) * (255 - i))) >> 8) & 0xf800) |
    (((((c & 0x07e0) * i) + ((d & 0x07e0) * (255 - i))) >> 8) & 0x07e0) |
    (((((c & 0x001f) * i) + ((d & 0x001f) * (255 - i))) >> 8) & 0x001f);
}

/*
** Displays the specified image on the framebuffer using specified technique
** NOTE: *ALL* conversion (bpp and size) must be done before displaying!
*/
int uimage_display_on_fb(uimage *image, int technique)
{
	uimage *cur_fb_contents;
	int fb_len, fb_fd;
	void *fb_mmap;
	
	/* Get our current framebuffer state */
	if ((cur_fb_contents = uimage_new()) == NULL)
		return -1;
	if (uimage_from_framebuffer(cur_fb_contents))
		return -2;
	if ((fb_len = get_framebuffer_length()) == 0)
		return -3;
	fb_fd = get_framebuffer_fd();
	
	/* If we aren't just fading out what is currently on the fb... */
	if ((image != NULL) || (technique != DISPTECH_FADEOUT)) {
		/* ...make sure our image fits */
		if (image->width != cur_fb_contents->width)
			return -4;
		if (image->height != cur_fb_contents->height)
			return -5;
		if (image->depth != cur_fb_contents->depth)
			return -6;
		if (image->stride != cur_fb_contents->stride)
		  {
		    logme("stride difference %d, %d\n", image->stride, cur_fb_contents->stride);
		    return -7;
		  }
	}
	
	/* Map the framebuffer to memory for display */
	fb_mmap = mmap(NULL, fb_len, PROT_WRITE, MAP_SHARED, fb_fd, 0);
	if (fb_mmap == MAP_FAILED)
		return -8;

	switch (technique)
	  {
	    /****************************************************
	     ****************    FILL IN BELOW    ****************
	     ****************************************************/

	    /* Do an immediate cut to the new image */
	  case DISPTECH_CUT:
	  DISPTECH_CUT:
	    /* Do we copy framebuffer length, framebuffer size or image length? */
	    /* stride * height  if they match */
	    memcpy(fb_mmap, image->data, image->height * image->stride);
	    break;
		
	    /* Fade to the new image from what is currently on the framebuffer */
	  case DISPTECH_CROSSFADE:
	    if(cur_fb_contents->depth == 8)
	      goto DISPTECH_CUT;
	    if(cur_fb_contents->depth == 16)
	    {
	      int fade, count, h, w, i;
	      unsigned long *dest, *a, *b;
	      count = image->width * image->height;

	      for(fade = 0; fade < 256; fade++)
		{
		  dest = fb_mmap;
		  a = cur_fb_contents->data;
		  b = image->data;
		  for(h = 0; h < image->height; h++)
		    {
		      for(w = 0; w < image->width; w+=2)
			/* this can be optimised heavily */
			*dest++ = mix16_fast2(*b++, *a++, fade);
		    }
		}
	    }
	    else
	      logme("uimage_display_on_fb(),  DISPTECH_CROSSFADE not implemented for %d bit colour", cur_fb_contents->depth);
	    
	    break;
		
		/* Black the screen first, then fade up from black */
	  case DISPTECH_FADEIN:
	    if(cur_fb_contents->depth == 16)
	    {
	      int fade, i;
	      unsigned long *dest, *b;
	      for(fade = 0; fade < 256; fade+=1)
		{
		  int i;
		  dest = fb_mmap;
		  b = image->data;
		  for(i = cur_fb_contents->height * cur_fb_contents->width; i > 0; i-=2)
		    {
		    /* this can be optimised heavily */
		    //*dest++ = fade16(*b++, fade);
		    //*dest++ = fade16_fast3(*b++, fade);
		    *dest++ = fade16_fast2(*b++, fade+1);
		    }
		}
	    }
	    else
	      logme("uimage_display_on_fb(),  DISPTECH_FADEIN not implemented for %d bit colour", cur_fb_contents->depth);
	    break;
		
		/* From whatever is currently on the screen, fade to black */
	  case DISPTECH_FADEOUT:
	    if(cur_fb_contents->depth == 8)
	      {
		/* palette magic fade */
	      }
	    else if(cur_fb_contents->depth == 16)
	      {
		int fade, i;
		unsigned long *dest, *a;
		for(fade = 255; fade >= 0; fade--)
		  {
		    int i;
		    dest = fb_mmap;
		    a = cur_fb_contents->data;
		    for(i = cur_fb_contents->height * cur_fb_contents->width; i > 0; i-=2)
		      /* this can be optimised heavily */
		      *dest++ = fade16_fast3(*a++, fade);
		  }
	      }
	    else if(cur_fb_contents->depth == 24)
	      {
		int fade, i;
		unsigned long *dest, *a;
		for(fade = 255; fade >= 0; fade--)
		  {
		    int i;
		    dest = fb_mmap;
		    a = cur_fb_contents->data;
		    for(i = cur_fb_contents->height * cur_fb_contents->width; i > 0; i-=2)
		      /* this can be optimised heavily */
		      *dest++ = fade16_fast3(*a++, fade);
		  }
	      }
	    else
	      logme("uimage_display_on_fb(),  DISPTECH_FADEOUT not implemented for %d bit colour", cur_fb_contents->depth);
	    break;
	  }
	
	/* Free up all our leftover stuff */
	close(fb_fd);
	munmap(fb_mmap, fb_len);
	uimage_free(cur_fb_contents);
	
	return 0;
}

static void generate_palette_16(unsigned short *dest, unsigned char *src)
{
  int i;
  int scale = 255;
  char c;
  for (i = 0; i < 256; i++, src += 3)
    {
    *dest++ =
      (((src[0]*scale)&0xf800) >> 0) |
      (((src[1]*scale)&0xfc00) >> 5) |
      (((src[2]*scale)&0xf800) >> 11);
    }
}

static unsigned short uimage_colour_24to16_scale(unsigned long colour)
{
  int scale = 255;
  return
    (((((colour>>16)&0xff)*scale)&0xf800) >> 0) |
    (((((colour>>8)&0xff)*scale)&0xfc00) >> 5) |
    (((((colour>>0)&0xff)*scale)&0xf800) >> 11);
}

static void convert8to16(unsigned short *dest, int count, unsigned char *src, unsigned short *pal)
{
  while(count--)
    *dest++ = pal[*src++];
}

/*
** Converts the specified uimage to the specified color depth
*/
int uimage_convert_depth(uimage *image, int bpp) 
{
  char *buffer;

  buffer = malloc(image->height * image->width * ((bpp+7)>>3));

  if (image->depth == bpp)
    return 0;
		
	/****************************************************
	****************    FILL IN BELOW    ****************
	****************************************************/
		
  switch (image->depth) {
  case 8:
    switch (bpp) {
    case 16:
      {
	unsigned short colour_map[256];

	generate_palette_16(colour_map, image->colourmap);

	{
	  int i;
	  int count = image->width*image->height;
	  unsigned char *src = image->data;
	  unsigned short *dest = (unsigned short *)buffer;

	  while(count-- > 0)
	    *dest++ = colour_map[*src++];
	}

	free(image->data);
	image->data = buffer;
	image->depth = bpp;
	free(image->colourmap);
	image->colourmap = NULL;
	image->stride = image->width * ((image->depth+7)>>3);
	return 0;
      }
      break;
    case 24:
      logme("uimage_convert_depth(8 -> 24) not implemented!\n");
      break;
    case 32:
      logme("uimage_convert_depth(8 -> 32) not implemented!\n");
      break;
    }
    break;
  case 16:
    switch (bpp) {
    case 8:
      logme("uimage_convert_depth(16 -> 8) not implemented!\n");
      break;
    case 24:
      logme("uimage_convert_depth(16 -> 24) not implemented!\n");
      break;
    case 32:
      logme("uimage_convert_depth(16 -> 32) not implemented!\n");
      break;
    }
    break;
  case 24:
    switch (bpp) {
    case 8:
      logme("uimage_convert_depth(24 -> 8) not implemented!\n");
      break;
    case 16:
      logme("uimage_convert_depth(24 -> 16) not implemented!\n");
      break;
    case 32:
      logme("uimage_convert_depth(24 -> 32) not implemented!\n");
      break;
    }
    break;
  case 32:
    switch (bpp) {
    case 8:
      logme("uimage_convert_depth(32 -> 8) not implemented!\n");
      break;
    case 16:
      logme("uimage_convert_depth(32 -> 16) not implemented!\n");
      break;
    case 24:
      logme("uimage_convert_depth(32 -> 24) not implemented!\n");
      break;
    }
    break;
  default:
    logme("uimage_convert_depth(%d -> %d) errnous perhaps?\n", image->depth, bpp);
    return -1;
  }
  return 0;
}

static void memset16(unsigned short *s, int c, int n)
{
  while(c-- > 0)
    *s++ = c;
}    

static void memset32(unsigned long *s, int c, int n)
{
  while(c-- > 0)
    *s++ = c;
}

static void set8(unsigned char *s, int c, int n)
{
  unsigned long *t = (unsigned long *)s;
  int a = (c << 24) | (c << 16) | (c << 8) | c;
  while(n > 3)
    {
      *t++ = a;
      n -= 4;
    }
  s = (unsigned char *)t;
  while(n-- > 0)
    *s++ = c;
}

static void set16(unsigned short *s, int c, int n)
{
  unsigned long *t = (unsigned long *)s;
  int a = c | (c << 16);
  while(n > 1)
    {
      *t++ = a;
      n -= 2;
    }
  if(n > 0)
    (unsigned short *)t = c;
}

static void set24(unsigned long *s, int c, int n)
{
  unsigned char *t;
  int r = c >> 16, g = (c >> 8) & 0xff, b = c & 0xff;
  int d = (c << 8) | r, e = (c << 16) | (c >> 16), f = (b << 24) | c;
  while(n > 3)
    {
      *s++ = d;
      *s++ = e;
      *s++ = f;
      n -= 4;
    }
  t = (unsigned char *)s;
  while(n-- > 0)
    {
      *t++ = r;
      *t++ = g;
      *t++ = b;
    }
}

static const inline int colour_24to16(int c)
{
  return
    ((c & 0xf80000) >> 8) |
    ((c & 0x00fc00) >> 5) |
    ((c & 0x0000f8) >> 3);
}

static void set32(unsigned long *s, int c, int n)
{
  while(n-- > 0)
    *s++ = c;
}

/*
** Centers a uimage in the size specified and fills the rest 
** from the primary palette color.
*/
int uimage_fill_to_size(uimage *image, int width, int height)
{
  int x0, x1, x2, x3;
  int y0, y1, y2, y3;
  int count, i, j;
  char *buffer, *buffer_saved, *s;
  int base_colour;

  if ((image->width == width) && (image->height == height))
    return 0;
	
  if (image->colourmap == NULL || image->depth != 8)
    return -1;

  if ((buffer = malloc(width * height)) == NULL)
    return -2;

  if (image->width > width || image->height > height)
    /* we should really handle this case (shrinking) */
    return -3;

  /* Center and fill */
  x0 = 0;
  y0 = 0;
  x1 = (width-image->width)/2;
  y1 = (height-image->height)/2;
  x2 = x1 + image->width;
  y2 = y1 + image->height;
  x3 = width;
  y3 = height;

  s = image->data;
  buffer_saved = buffer;
  base_colour = *(char *)image->data;

  count = (y1-y0)*(x3-x0);
  memset(buffer, base_colour, count);
  buffer += count;

  for(j = y1; j < y2; j++)
    {
      memset(buffer + x0, base_colour, x1 - x0);
      memcpy(buffer + x1, s, x2 - x1);
      memset(buffer + x2, base_colour, x3 - x2);
      s += x2 - x1;
      buffer += x3 - x0;
    }

  count = (y3-y2)*(x3-x0);
  memset(buffer, base_colour, count);
  buffer += count;

  buffer = buffer_saved;

  free(image->data);
  image->data = buffer;
  image->width = width;
  image->height = height;
  image->stride = width;

  return 0;
}

int uimage_display_icon(uimage *image, int x, int y)
{
  int x0, x1, x2, x3, xa;
  int y0, y1, y2, y3;
  int count, i, j;
  char *buffer, *buffer_saved;
  int base_colour;
  unsigned short *s, *d;
  unsigned long border = 0x000000, left = 0xff9999, right = 0x9999ff, background = 0xffffff;

  uimage *cur_fb_contents;
  int fb_len, fb_fd;
  //  void *fb_mmap;

  /* Get our current framebuffer state */
  if ((cur_fb_contents = uimage_new()) == NULL)
    return -1;
  if (uimage_from_framebuffer_fast(cur_fb_contents))
    return -2;
  if ((fb_len = get_framebuffer_length()) == 0)
    return -3;
  fb_fd = get_framebuffer_fd();

  if (image->depth != cur_fb_contents->depth)
    return -1;
  
  /* Map the framebuffer to memory for display */
  if(fb_mmap == NULL)
    {
      fb_mmap = mmap(NULL, fb_len, PROT_WRITE, MAP_SHARED, fb_fd, 0);
      if (fb_mmap == MAP_FAILED)
	return -8;
    }

  /* Center and fill */
  x0 = 0;
  y0 = 0;
  x1 = x;
  y1 = y;
  x2 = x1 + image->width;
  y2 = y1 + image->height;
  x3 = cur_fb_contents->width;
  y3 = cur_fb_contents->height;

  d = (unsigned short *)(fb_mmap + y1 * cur_fb_contents->stride);
  s = (unsigned short *)image->data;

  for(j = 0; j < image->height; j++)
    {
      memcpy(d + x1, s, image->stride);
      d += x3 - x0;
      s += image->width;
    }

  //  munmap(fb_mmap, fb_len);
  close(fb_fd);
  uimage_free(cur_fb_contents);

  return 0;
}

int uimage_display_progress(int percentage)
{
  int x0, x1, x2, x3, xa;
  int y0, y1, y2, y3;
  int count, i, j;
  char *buffer, *buffer_saved;
  int base_colour;
  unsigned short *s;
  unsigned long border = 0x000000, left = 0xff9999, right = 0x9999ff, background = 0xffffff;

  int width = 500, height = 24;

  uimage *cur_fb_contents;
  int fb_len, fb_fd;
  //void *fb_mmap;

  /* Get our current framebuffer state */
  if ((cur_fb_contents = uimage_new()) == NULL)
    return -1;
  if (uimage_from_framebuffer_fast(cur_fb_contents))
    return -2;
  if ((fb_len = get_framebuffer_length()) == 0)
    return -3;
  fb_fd = get_framebuffer_fd();

  /* Map the framebuffer to memory for display */
  if(fb_mmap == NULL)
    {
      fb_mmap = mmap(NULL, fb_len, PROT_WRITE, MAP_SHARED, fb_fd, 0);
      if (fb_mmap == MAP_FAILED)
	return -8;
    }

  if(percentage >= 99)
    border = left = right = background;

  /* Center and fill */
  x0 = 0;
  y0 = 0;
  x1 = (cur_fb_contents->width-width)/2;
  y1 = 20;
  x2 = x1 + width;
  y2 = y1 + height;
  x3 = cur_fb_contents->width;
  y3 = cur_fb_contents->height;

  xa = (((x2 - x1 - 4) * percentage) / 100 + x1 + 2) & ~1;

  //s = (unsigned short *)(fb_mmap + y1 * cur_fb_contents->stride);

  s = (unsigned short *)(fb_mmap + y1 * cur_fb_contents->stride);

  set16(s + x1, colour_24to16(border), x2 - x1);
  s += x3 - x0;
  set16(s + x1, colour_24to16(border), x2 - x1);
  s += x3 - x0;

  for(j = (y1 + 2); j < (y2 - 2); j++)
    {
      set16(s + x1,     colour_24to16(border), 2);
      set16(s + x1 + 2, colour_24to16(left), xa - x1 - 2);
      set16(s + xa, colour_24to16(right), x2 - xa - 1);
      set16(s + x2 - 2, colour_24to16(border), 2);
      s += x3 - x0;
    }

  set16(s + x1, colour_24to16(border), x2 - x1);
  s += x3 - x0;
  set16(s + x1, colour_24to16(border), x2 - x1);
  s += x3 - x0;

  //  set16(fb_mmap, 0, 1024*100);

  //munmap(fb_mmap, fb_len);
  close(fb_fd);
  uimage_free(cur_fb_contents);

  return 0;
}

